#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python Team Awareness Kit (PyTAK) Module Constants."""

import logging
import os

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2020 Orion Labs, Inc."
__license__ = "Apache License, Version 2.0"


if bool(os.environ.get('DEBUG')):
    LOG_LEVEL = logging.DEBUG
    LOG_FORMAT = logging.Formatter(
        ('%(asctime)s pytak %(levelname)s %(name)s.%(funcName)s:%(lineno)d - '
         '%(message)s'))
    logging.debug('pytak Debugging Enabled via DEBUG Environment Variable.')
else:
    LOG_LEVEL = logging.INFO
    LOG_FORMAT = logging.Formatter(
        ('%(asctime)s pytak %(levelname)s - %(message)s'))


DEFAULT_COT_PORT: int = 8087
DEFAULT_ATAK_PORT: int = 4242
DEFAULT_BROADCAST_PORT: int = 6969
DEFAULT_BACKOFF: int = 120
DEFAULT_SLEEP: int = 5

DEFAULT_HEX_RANGES: dict = {
    "US-CIV": {"start": 0xA00000, "end": 0xAFFFFF},
    "CA-CIV": {"start": 0xC00000, "end": 0xC3FFFF},
    "NZ-CIV": {"start": 0xC80000, "end": 0xC87FFF},
    "AU-CIV": {"start": 0x440000, "end": 0x447FFF},
    "UK-CIV": {"start": 0x400000, "end": 0x43FFFF}
}

DOMESTIC_AIRLINES: list = [
    "AAL",
    "UAL",
    "FDX",
    "UPS",
    "SWA"
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Python Team Awareness Kit (PyTAK) Module.

"""
Python Team Awareness Kit (PyTAK) Module.
~~~~


:author: Greg Albrecht W2GMD <oss@undef.net>
:copyright: Copyright 2020 Orion Labs, Inc.
:license: Apache License, Version 2.0
:source: <https://github.com/ampledata/pytak>

"""

from .constants import (LOG_LEVEL, LOG_FORMAT, DEFAULT_COT_PORT,  # NOQA
                        DEFAULT_BACKOFF, DEFAULT_SLEEP,
                        DEFAULT_ATAK_PORT, DEFAULT_BROADCAST_PORT,
                        DOMESTIC_AIRLINES, DEFAULT_HEX_RANGES)

from .functions import (split_host, udp_client, parse_cot_url,  # NOQA
                        faa_to_cot_type, multicast_client, eventworker_factory)

from .classes import (NetworkClient, CoTWorker, AsyncNetworkClient,  # NOQA
                      AsyncCoTWorker, EventWorker, MessageWorker)

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2020 Orion Labs, Inc."
__license__ = "Apache License, Version 2.0"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python Team Awareness Kit (PyTAK) Module Class Definitions."""

import asyncio
import logging
import os
import queue
import random
import socket
import threading
import time
import urllib

import pytak
import pycot

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2020 Orion Labs, Inc."
__license__ = "Apache License, Version 2.0"


# Dear Reader, Py3 doesn't need to inherit from Object anymore!
class NetworkClient:
    """CoT Network Client (TX)."""

    _logger = logging.getLogger(__name__)
    if not _logger.handlers:
        _logger.setLevel(pytak.LOG_LEVEL)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(pytak.LOG_LEVEL)
        _console_handler.setFormatter(pytak.LOG_FORMAT)
        _logger.addHandler(_console_handler)
        _logger.propagate = False

    def __init__(self, cot_host: str, cot_port: int = None,
                 broadcast: bool = False) -> None:
        self.broadcast = broadcast

        self.socket: socket.socket = None
        self.addr: str = None
        self.port: int = None

        if ':' in cot_host:
            self.addr, port = cot_host.split(':')
            self.port = int(port)
        elif cot_port:
            self.addr = cot_host
            self.port = int(cot_port)
        else:
            self.addr = cot_host
            self.port = int(pytak.DEFAULT_COT_PORT)

        self.socket_addr = f'{self.addr}:{self.port}'

        if self.broadcast:
            self._logger.info(
                'Using Broadcast Socket, CoT Destination: %s',
                self.socket_addr)
            self._setup_broadcast_socket()
        else:
            self._logger.info(
                'Using Unicast Socket, CoT Destination: %s',
                self.socket_addr)
            self._setup_unicast_socket()

    def _setup_unicast_socket(self) -> None:
        """Sets up the TCP Unicast Socket for sending CoT events."""
        self._logger.debug(
            'Setting up Unicast Socket to CoT Destination: %s',
            self.socket_addr)
        if self.socket is not None:
            self.socket.close()
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.addr, self.port))

    def _setup_broadcast_socket(self) -> None:
        """Sets up the UDP Broadcast Socket for sending CoT events."""
        self._logger.debug(
            'Setting up Broadcast Socket to CoT Destination: %s',
            self.socket_addr)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    def send_cot(self, event: bytes, timeout: int = 10) -> bool:
        """Wrapper for sending TCP Unicast or UDP Broadcast CoT Events."""
        if os.environ.get('DONT_ADD_NEWLINE'):
            _event = event
        else:
            _event = event + b'\n'

        self._logger.debug('Sending CoT to %s: "%s"', self.socket_addr, _event)

        if self.broadcast:  # pylint: disable=no-else-return
            return self.sendto(_event)
        else:
            return self.sendall(_event, timeout)

    def close(self):
        """Closes this instance's network socket."""
        return self.socket.close()

    def sendall(self, event: bytes, timeout: int = 10) -> bool:
        """Sends a CoT Event to a TCP Unicast address."""
        # is the socket alive?
        if self.socket.fileno() is -1:
            self._logger.warning(
                'Restarting Socket as socket.fileno() returned -1')
            self._setup_unicast_socket()
            return False

        self.socket.settimeout(timeout)

        try:
            self.socket.sendall(event)
            if not os.environ.get('DISABLE_RANDOM_SLEEP'):
                time.sleep(pytak.DEFAULT_SLEEP * random.random())
            return True
        except Exception as exc:
            self._logger.error(
                'socket.sendall() raised an Exception, sleeping: ')
            self._logger.exception(exc)
            time.sleep(pytak.DEFAULT_BACKOFF * random.random())
            self._setup_unicast_socket()
            return False

    def sendto(self, event: bytes) -> bool:
        """Sends a CoT Event to a UDP Broadcast address."""
        try:
            self.socket.sendto(event, (self.addr, self.port))
            return True
        except Exception as exc:
            self._logger.error(
                'socket.sendto() raised an Exception, sleeping: ')
            self._logger.exception(exc)
            time.sleep(pytak.DEFAULT_BACKOFF * random.random())
            self._setup_broadcast_socket()
            return False


class CoTWorker(threading.Thread):

    """CoTWorker Thread."""

    _logger = logging.getLogger(__name__)
    if not _logger.handlers:
        _logger.setLevel(pytak.LOG_LEVEL)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(pytak.LOG_LEVEL)
        _console_handler.setFormatter(pytak.LOG_FORMAT)
        _logger.addHandler(_console_handler)
        _logger.propagate = False

    def __init__(self, msg_queue: queue.Queue, cot_host: str,
                 cot_port: int = None, broadcast: bool = False) -> None:
        self.msg_queue: queue.Queue = msg_queue

        self.net_client = NetworkClient(
            cot_host=cot_host,
            cot_port=cot_port,
            broadcast=broadcast
        )

        # Thread setup:
        threading.Thread.__init__(self)
        self.daemon = True
        self._stopper = threading.Event()

    def stop(self):
        """Stop the thread at the next opportunity."""
        self._logger.debug('Stopping CoTWorker')
        self.net_client.close()
        self._stopper.set()

    def stopped(self):
        """Checks if the thread is stopped."""
        return self._stopper.isSet()

    def run(self):
        """Runs this Thread, reads in Message Queue & sends out CoT."""
        self._logger.info('Running CoTWorker')

        while not self.stopped():
            try:
                msg = self.msg_queue.get(True, 1)
                self._logger.debug('From msg_queue: "%s"', msg)
                if not msg:
                    continue
                self.net_client.send_cot(msg)
            except queue.Empty:
                recv = self.net_client.socket.recv(1024)
                pass


class AsyncNetworkClient(asyncio.Protocol):
    """Async CoT Network Client (TX)."""

    _logger = logging.getLogger(__name__)
    if not _logger.handlers:
        _logger.setLevel(pytak.LOG_LEVEL)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(pytak.LOG_LEVEL)
        _console_handler.setFormatter(pytak.LOG_FORMAT)
        _logger.addHandler(_console_handler)
        _logger.propagate = False
    logging.getLogger('asyncio').setLevel(pytak.LOG_LEVEL)

    def __init__(self, msg_queue: queue.Queue, on_con_lost) -> None:
        self.transport = None
        self.msg_queue = msg_queue
        self.on_con_lost = on_con_lost

    def connection_made(self, transport):
        self.transport = transport
        self.address = transport.get_extra_info('peername')
        self._logger.debug('Connected to %s', self.address)

    def connection_lost(self, exc):
        self._logger.warning('Disconnected from %s', self.address)
        self._logger.exception(exc)
        self.on_con_lost.set_result(None)

    def error_received(self, exc):
        self._logger.warning('Disconnected from %s', self.address)
        self._logger.exception(exc)
        #if not self.on_con_lost.done():
        self.on_con_lost.set_result(None)

    def eof_received(self):
        self._logger.debug('Received EOF from %s', self.address)
        self.transport.close()
        #if not self.on_con_lost.done():
        self.on_con_lost.set_result(None)

    def wait_connection_lost(self):
        return self.on_con_lost.done()


class AsyncCoTWorker:

    """Async CoTWorker."""

    _logger = logging.getLogger(__name__)
    if not _logger.handlers:
        _logger.setLevel(pytak.LOG_LEVEL)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(pytak.LOG_LEVEL)
        _console_handler.setFormatter(pytak.LOG_FORMAT)
        _logger.addHandler(_console_handler)
        _logger.propagate = False
    logging.getLogger('asyncio').setLevel(pytak.LOG_LEVEL)

    def __init__(self, msg_queue: asyncio.Queue, transport) -> None:
        self.msg_queue = msg_queue
        self.transport = transport

    async def run(self):
        """Runs this Thread, reads in Message Queue & sends out CoT."""
        self._logger.info('Running AsyncCoTWorker')

        while 1:
            msg = await self.msg_queue.get()
            self._logger.debug('From msg_queue: "%s"', msg)
            if not msg:
                continue
            self.transport.write(msg)
            if not os.environ.get('DISABLE_RANDOM_SLEEP'):
                await asyncio.sleep(pytak.DEFAULT_SLEEP * random.random())


class EventWorker:

    """EventWorker."""

    _logger = logging.getLogger(__name__)
    if not _logger.handlers:
        _logger.setLevel(pytak.LOG_LEVEL)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(pytak.LOG_LEVEL)
        _console_handler.setFormatter(pytak.LOG_FORMAT)
        _logger.addHandler(_console_handler)
        _logger.propagate = False
    logging.getLogger('asyncio').setLevel(pytak.LOG_LEVEL)

    def __init__(self, event_queue: asyncio.Queue, writer) -> None:
        self.event_queue = event_queue
        self.writer = writer

    async def run(self):
        """Runs this Thread, reads in Message Queue & sends out CoT."""
        self._logger.info('Running EventWorker')

        while 1:
            event = await self.event_queue.get()
            if not event:
                continue
            self._logger.debug("event='%s'", event)

            if isinstance(event, pycot.Event):
                _event = event.render(encoding='UTF-8', standalone=True)
            else:
                _event = event

            if hasattr(self.writer, "send"):
                await self.writer.send(_event)
            else:
                self.writer.write(_event)
                await self.writer.drain()
            if not os.environ.get('DISABLE_RANDOM_SLEEP'):
                await asyncio.sleep(pytak.DEFAULT_SLEEP * random.random())


class MessageWorker:

    _logger = logging.getLogger(__name__)
    if not _logger.handlers:
        _logger.setLevel(pytak.LOG_LEVEL)
        _console_handler = logging.StreamHandler()
        _console_handler.setLevel(pytak.LOG_LEVEL)
        _console_handler.setFormatter(pytak.LOG_FORMAT)
        _logger.addHandler(_console_handler)
        _logger.propagate = False
    logging.getLogger('asyncio').setLevel(pytak.LOG_LEVEL)

    def __init__(self, event_queue: asyncio.Queue,
                 url: urllib.parse.ParseResult, cot_stale: int = None):
        self.event_queue = event_queue
        self.url = url
        self.cot_stale = cot_stale

    async def _put_event_queue(self, event: pycot.Event) -> None:
        """Puts Event onto the CoT Transmit Queue."""
        try:
            await self.event_queue.put(event)
        except queue.Full as exc:
            self._logger.warning(
                'Lost CoT Event (queue full): "%s"', _event)

    async def run(self) -> None:
        self._logger.info("Overwrite this method!")
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Setup for the Python Team Awareness Kit (PyTAK) Module.

:author: Greg Albrecht W2GMD <oss@undef.net>
:copyright: Copyright 2020 Orion Labs, Inc.
:license: Apache License, Version 2.0
:source: <https://github.com/ampledata/pytak>
"""

import os
import setuptools
import sys

__title__ = "pytak"
__version__ = "3.0.0b3"
__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2020 Orion Labs, Inc."
__license__ = "Apache License, Version 2.0"


def publish():
    """Function for publishing package to pypi."""
    if sys.argv[-1] == "publish":
        os.system("python setup.py sdist")
        os.system("twine upload dist/*")
        sys.exit()


publish()


setuptools.setup(
    version=__version__,
    name=__title__,
    packages=[__title__],
    package_dir={__title__: __title__},
    url=f"https://github.com/ampledata/{__title__}",
    description="Python Team Awareness Kit (PyTAK) Module",
    author="Greg Albrecht",
    author_email="oss@undef.net",
    package_data={"": ["LICENSE"]},
    license=open("LICENSE").read(),
    long_description=open("README.rst").read(),
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License"
    ],
    keywords=[
        "Cursor on Target", "ATAK"
    ],
    install_requires=[
        "asyncio_dgram",
        "pycot"
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""APRS Cursor-on-Target Constants."""

import asyncio
import socket
import struct

import asyncio_dgram

import pytak

__author__ = "Greg Albrecht W2GMD <oss@undef.net>"
__copyright__ = "Copyright 2020 Orion Labs, Inc."
__license__ = "Apache License, Version 2.0"


def split_host(host, port) -> tuple:
    if ':' in host:
        addr, port = host.split(':')
        port = int(port)
    elif port:
        addr = host
        port = int(port)
    else:
        addr = host
        port = int(pytak.DEFAULT_COT_PORT)

    return addr, port


def parse_cot_url(url) -> tuple:
    if ":" in url.path:
        host, port = str(url.path).split(":")
    else:
        host = url.path
        if "broadcast" in url.scheme:
            port = pytak.DEFAULT_BROADCAST_PORT
        else:
            port = pytak.DEFAULT_COT_PORT
    return host, port


async def udp_client(url):
    """Create a UDP Network client, simulate other transports."""
    host, port = parse_cot_url(url)
    stream = await asyncio_dgram.connect((host, port))
    if "broadcast" in url.scheme:
        sock = stream.socket
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    return stream


async def multicast_client(url):
    """Create a UDP Network client, simulate other transports."""
    host, port = parse_cot_url(url)
    group = socket.inet_aton(host)
    mreq = struct.pack('4sL', group, socket.INADDR_ANY)
    stream = await asyncio_dgram.bind((host, port))
    sock = stream.socket
    #sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    return stream


async def eventworker_factory(cot_url: str, event_queue,
                              fts_token: str = None):
    """
    Creates a Cursor on Target Event Worker based on URL parameters.

    :param cot_url: URL to CoT Destination.
    :param event_queue: asyncio.Queue worker to get events from.
    :param fts_token: If supplied, API Token to use for FreeTAKServer REST.
    :return: EventWorker or asyncio Protocol
    """
    # CoT/TAK Event Workers (transmitters):
    if "http" in cot_url.scheme and fts_token:
        eventworker = pytak.FTSClient(
            event_queue,
            cot_url.geturl(),
            fts_token
        )
    elif "tcp" in cot_url.scheme:
        host, port = pytak.parse_cot_url(cot_url)
        _, writer = await asyncio.open_connection(host, port)
        eventworker = pytak.EventWorker(event_queue, writer)
    elif "udp" in cot_url.scheme:
        writer = await pytak.udp_client(cot_url)
        eventworker = pytak.EventWorker(event_queue, writer)

    return eventworker


def faa_to_cot_type(icao_hex: int, category: str = None,
                    flight: str = None) -> str:
    """
    Classify Cursor on Target Event Type from ICAO, and if available, from
    Emitter Category & Flight.
    """
    cm = "C"  # Civ
    attitude = "u"  # Unknown

    icao_int = int(f"0x{icao_hex.replace('~', '')}", 16)

    if flight:
        for dom in pytak.DOMESTIC_AIRLINES:
            if flight.startswith(dom):
                # SN: Should be "n" for Mil/Fed posture.
                attitude = "f"  # FIXME: Default posture depends on user.

    # SN: Leave "neutral" as Taiwan uses this subset instead of PRC China....
    tw_start = 0x899000
    tw_end = 0x8993FF
    if tw_start <= icao_int <= tw_end:
        attitude = "n"

    # SN: Make the US range (all non-mil) ICAO addresses just generic "n"
    us_civ_start = pytak.DEFAULT_HEX_RANGES["US-CIV"]["start"]
    us_civ_end = pytak.DEFAULT_HEX_RANGES["US-CIV"]["end"]
    if us_civ_start <= icao_int <= us_civ_end:
        attitude = "n"

    # Friendly Mil:
    # mil = ["US-MIL", "UK-MIL", "CA-MIL", "NZ-MIL", "AU-MIL"]
    # for fvey in mil:
    #    mil_start = pytak.DEFAULT_HEX_RANGES[fvey]["start"]
    #    mil_end = pytak.DEFAULT_HEX_RANGES[fvey]["end"]
    #    if mil_start <= icao_int <= mil_end:
    #        attitude = "f"
    #        cm = "M"

    # Default Fixed Wing
    cot_type = f"a-{attitude}-A-{cm}-F"

    if category:
        _category = str(category)

        if _category in ["7", "A7"]:  # Rotor/Helicopter
            cot_type = f"a-{attitude}-A-{cm}-H"
        if _category in ["10", "B2"]:  # Balloon
            cot_type = f"a-{attitude}-A-{cm}-L"
        elif _category in ["14", "B6"]:  # Drone
            cot_type = f"a-{attitude}-A-{cm}-F-q"
        elif _category in ["17", "18", "C1", "C2"]:
            cot_type = f"a-.-G-E-V-C-U"
        elif _category in ["19"]:
            cot_type = f"a-{attitude}-G-I-U-T-com-tow"

    return cot_type
